from abc import ABC, abstractmethod
from .polymer_classifier import PolymerClassifier
import random
import os
from aldepyde.data.distribution import Distribution, load, available_distributions
from aldepyde.data.distributions.standards import distribution_head
import json

class ResidueGenerator(ABC):
    def __init__(self, classifier:PolymerClassifier):
        self.classifier = classifier

    def _generate(self):
        pass

    # Just pure, random generation. Nothing fancy
    def random(self, length:int, n:int=1) -> list:
        alphabet = list(self.classifier.alphabet)
        sequences = []
        for i in range(n):
            sequences.append("".join(random.choices(alphabet, k=length)))
        return sequences

    def random_from_distribution(self, distribution: os.PathLike|str|dict[str, float]|Distribution, length:int|tuple[int,int], n:int=1, exclude:tuple=()) -> list:
        if isinstance(distribution, str) and os.path.isfile(distribution): # Given a file
            with open(distribution) as fp:
                data = json.load(fp)
                if distribution_head not in data.keys():
                    raise ValueError(f"Key '{distribution_head}' must be in the top level of your provided file")
            usable_distribution = Distribution(None, data[distribution_head])
        elif isinstance(distribution, str) and not os.path.isfile(distribution): # Given a preload
            usable_distribution = load(distribution)
        elif isinstance(distribution, Distribution): # Given a distribution
            usable_distribution = distribution
        elif isinstance(distribution, dict): # Given a dictionary
            usable_distribution = Distribution(None, distribution)
        else:
            raise ValueError('Invalid distribution')
        for r in exclude:
            usable_distribution.eliminate_entry(r)
        if not isinstance(length, int):
            size = lambda: random.randint(length[0], length[1])
        else:
            size = lambda: length
        usable_distribution.normalize_map()
        sequences = []
        for _ in range(n):
            sequence = random.choices(population=list(usable_distribution.frequency_map.keys()), weights=list(usable_distribution.frequency_map.values()), k=size())
            sequences.append("".join(sequence))
        return sequences




from typing import List, Tuple, Optional, Dict
from dataclasses import dataclass, field


@dataclass
class Atom():
    serial: str
    name: str
    coord: Tuple[float, float, float]
    occupancy: float
    b_factor: float
    is_het: bool
    element: Optional[str]
    charge: Optional[str]
    altloc: Optional[str]


@dataclass
class Residue():
    name: str
    id: str
    ins_code: Optional[str]
    atoms: List[Atom] = field(default_factory=list)

    def add_atom(self, atom: Atom) -> None:
        self.atoms.append(atom)

@dataclass
class Chain():
    id: str
    residues: Dict[str, Residue] = field(default_factory=dict)

    def add_residue(self, residue: Residue):
        self.residues[residue.id] = residue
        # self.residues.append(residue)

    def get_residue(self, id):
        return self.residues[id]

@dataclass
class Model:
    id: str
    chains: Dict[str, Chain] = field(default_factory=dict)

    def add_chain(self, chain: Chain) -> None:
        self.chains[chain.id] = chain
        # self.chains.append(chain)

    def get_chain(self, id) -> Chain:
        return self.chains[id]

@dataclass
class Structure:
    id: str
    # title: str = ""
    models: Dict[str, Model] = field(default_factory=dict)

    def add_model(self, model: Model) -> None:
        self.models[model.id] = model
        # self.models.append(model)


from .polymer_classifier import PolymerClassifier

class ProteinClassifier(PolymerClassifier):
    DEFAULT_ALPHABET = set("ACDEFGHIKLMNPQRSTVWY")
    DEFAULT_CLASSES = classes = {
                "positive" : set("RHK"),
                "negative" : set("ED"),
                "polar" : set("NSCTQ"),
                "nonpolar" : set("MALIVGP"),
                "aromatic" : set("FYW"),
                "_other" : set("")
            }
    DEFAULT_FLAGS = {
                "handle_non_canonical" : True
            }

    #TODO Current conflict exists between custom classes and flags
    #What do you do if you create a custom class and want to generate with it?
    #What if a flag combines classes, or enables behavior such as handle_non_canonical?

    def __init__(self, alphabet: str | list | set = None, classes:dict = None, flags:dict|None = None):
        if alphabet is None:
            alphabet = self.DEFAULT_ALPHABET
        if classes is None:
            classes = self.DEFAULT_CLASSES
        if flags is None:
            flags = self.DEFAULT_FLAGS
        super().__init__(alphabet, classes, flags)

    def default_alphabet(self):
        self.alphabet = self.DEFAULT_ALPHABET

    def default_classes(self):
        self.classes = self.DEFAULT_CLASSES

    def default_flags(self):
        self.flags = self.DEFAULT_FLAGS




